@@ -8,6 +8,7 @@ from . import (
     means,
     mlls,
     models,
+    optim,
     priors,
     settings,
     utils,
@@ -30,7 +31,7 @@ from .lazy import cat, delazify, lazify
 from .mlls import ExactMarginalLogLikelihood
 from .module import Module
 
-__version__ = "1.1.1"
+__version__ = "1.2.1"
 
 __all__ = [
     # Submodules
@@ -41,6 +42,7 @@ __all__ = [
     "means",
     "mlls",
     "models",
+    "optim",
     "priors",
     "utils",
     "variational",
