@@ -2,7 +2,8 @@
 
 import torch
 
-from ..lazy import BlockDiagLazyTensor, BlockInterleavedLazyTensor, CatLazyTensor, LazyTensor
+from ..lazy import BlockDiagLazyTensor, BlockInterleavedLazyTensor, CatLazyTensor, LazyTensor, lazify
+from ..utils.broadcasting import _mul_broadcast_shape
 from .multivariate_normal import MultivariateNormal
 
 
@@ -33,6 +34,25 @@ class MultitaskMultivariateNormal(MultivariateNormal):
         if mean.dim() < 2:
             raise RuntimeError("mean should be a matrix or a batch matrix (batch mode)")
 
+        # Ensure that shapes are broadcasted appropriately across the mean and covariance
+        # Means can have singleton dimensions for either the `n` or `t` dimensions
+        batch_shape = _mul_broadcast_shape(mean.shape[:-2], covariance_matrix.shape[:-2])
+        if mean.shape[-2:].numel() != covariance_matrix.size(-1):
+            if covariance_matrix.size(-1) % mean.shape[-2:].numel():
+                raise RuntimeError(
+                    f"mean shape {mean.shape} is incompatible with covariance shape {covariance_matrix.shape}"
+                )
+            elif mean.size(-2) == 1:
+                mean = mean.expand(*batch_shape, covariance_matrix.size(-1) // mean.size(-1), mean.size(-1))
+            elif mean.size(-1) == 1:
+                mean = mean.expand(*batch_shape, mean.size(-2), covariance_matrix.size(-2) // mean.size(-2))
+            else:
+                raise RuntimeError(
+                    f"mean shape {mean.shape} is incompatible with covariance shape {covariance_matrix.shape}"
+                )
+        else:
+            mean = mean.expand(*batch_shape, *mean.shape[-2:])
+
         self._output_shape = mean.shape
         # TODO: Instead of transpose / view operations, use a PermutationLazyTensor (see #539) to handle interleaving
         self._interleaved = interleaved
@@ -214,6 +234,25 @@ class MultitaskMultivariateNormal(MultivariateNormal):
             return samples.view(new_shape).transpose(-1, -2).contiguous()
         return samples.view(sample_shape + self._output_shape)
 
+    def to_data_independent_dist(self):
+        """
+        Convert a multitask MVN into a batched (non-multitask) MVNs
+        The result retains the intertask covariances, but gets rid of the inter-data covariances.
+        The resulting distribution will have :attr:`len(mvns)` tasks, and the tasks will be independent.
+
+        :returns: the bached data-independent MVN
+        :rtype: gpytorch.distributions.MultivariateNormal
+        """
+        # Create batch distribution where all data are independent, but the tasks are dependent
+        full_covar = self.lazy_covariance_matrix
+        num_data, num_tasks = self.mean.shape[-2:]
+        data_indices = torch.arange(0, num_data * num_tasks, num_tasks, device=full_covar.device).view(-1, 1, 1)
+        task_indices = torch.arange(num_tasks, device=full_covar.device)
+        task_covars = full_covar[
+            ..., data_indices + task_indices.unsqueeze(-2), data_indices + task_indices.unsqueeze(-1)
+        ]
+        return MultivariateNormal(self.mean, lazify(task_covars).add_jitter())
+
     @property
     def variance(self):
         var = super().variance
