@@ -3,8 +3,10 @@
 import math
 import warnings
 from abc import ABC, abstractmethod
+from typing import Optional, Tuple
 
 import torch
+from torch import Tensor
 
 import gpytorch
 
@@ -14,11 +16,14 @@ from ..functions._inv_quad import InvQuad
 from ..functions._inv_quad_log_det import InvQuadLogDet
 from ..functions._matmul import Matmul
 from ..functions._root_decomposition import RootDecomposition
+from ..functions._sqrt_inv_matmul import SqrtInvMatmul
 from ..utils.broadcasting import _matmul_broadcast_shape, _mul_broadcast_shape
 from ..utils.cholesky import psd_safe_cholesky
 from ..utils.deprecation import _deprecate_renamed_methods
+from ..utils.errors import CachingError
 from ..utils.getitem import _compute_getitem_size, _convert_indices_to_tensors, _is_noop_index, _noop_index
-from ..utils.memoize import add_to_cache, cached
+from ..utils.memoize import add_to_cache, cached, pop_from_cache
+from ..utils.pivoted_cholesky import pivoted_cholesky
 from ..utils.warnings import NumericalWarning
 from .lazy_tensor_representation_tree import LazyTensorRepresentationTree
 
@@ -262,7 +267,7 @@ class LazyTensor(ABC):
             This method is used internally by the related function :func:`~gpytorch.lazy.LazyTensor.expand`,
             which does some additional work. Calling this method directly is discouraged.
         """
-        current_shape = torch.Size([1 for _ in range(len(batch_shape) - self.dim() - 2)] + list(self.batch_shape))
+        current_shape = torch.Size([1 for _ in range(len(batch_shape) - self.dim() + 2)] + list(self.batch_shape))
         batch_repeat = torch.Size(
             [expand_size // current_size for expand_size, current_size in zip(batch_shape, current_shape)]
         )
@@ -386,7 +391,7 @@ class LazyTensor(ABC):
         return self.diag()
 
     @cached(name="cholesky")
-    def _cholesky(self):
+    def _cholesky(self, upper=False):
         """
         (Optional) Cholesky-factorizes the LazyTensor
 
@@ -394,9 +399,9 @@ class LazyTensor(ABC):
             This method is used as an internal helper. Calling this method directly is discouraged.
 
         Returns:
-            (LazyTensor) Cholesky factor
+            (TriangularLazyTensor) Cholesky factor
         """
-        from .non_lazy_tensor import NonLazyTensor
+        from .triangular_lazy_tensor import TriangularLazyTensor
         from .keops_lazy_tensor import KeOpsLazyTensor
 
         evaluated_kern_mat = self.evaluate_kernel()
@@ -408,13 +413,13 @@ class LazyTensor(ABC):
 
         # if the tensor is a scalar, we can just take the square root
         if evaluated_mat.size(-1) == 1:
-            return NonLazyTensor(evaluated_mat.clamp_min(0.0).sqrt())
+            return TriangularLazyTensor(evaluated_mat.clamp_min(0.0).sqrt())
 
         # contiguous call is necessary here
-        cholesky = psd_safe_cholesky(evaluated_mat).contiguous()
-        return NonLazyTensor(cholesky)
+        cholesky = psd_safe_cholesky(evaluated_mat, jitter=settings.cholesky_jitter.value(), upper=upper).contiguous()
+        return TriangularLazyTensor(cholesky, upper=upper)
 
-    def _cholesky_solve(self, rhs):
+    def _cholesky_solve(self, rhs, upper: bool = False):
         """
         (Optional) Assuming that `self` is a Cholesky factor, computes the cholesky solve
 
@@ -424,7 +429,7 @@ class LazyTensor(ABC):
         Returns:
             (LazyTensor) Cholesky factor
         """
-        return torch.cholesky_solve(rhs, self.evaluate())
+        raise NotImplementedError("_cholesky_solve not implemented for the base LazyTensor")
 
     def _inv_matmul_preconditioner(self):
         """
@@ -685,22 +690,28 @@ class LazyTensor(ABC):
         Args:
             - diag (Scalar Tensor)
         """
-        from .diag_lazy_tensor import DiagLazyTensor
+        from .diag_lazy_tensor import ConstantDiagLazyTensor, DiagLazyTensor
         from .added_diag_lazy_tensor import AddedDiagLazyTensor
 
         if not self.is_square:
             raise RuntimeError("add_diag only defined for square matrices")
 
-        try:
-            expanded_diag = diag.expand(self.shape[:-1])
-        except RuntimeError:
-            raise RuntimeError(
-                "add_diag for LazyTensor of size {} received invalid diagonal of size {}.".format(
-                    self.shape, diag.shape
+        diag_shape = diag.shape
+        if len(diag_shape) == 0 or diag_shape[-1] == 1:
+            # interpret scalar tensor or single-trailing element as constant diag
+            diag_tensor = ConstantDiagLazyTensor(diag, diag_shape=self.shape[-1])
+        else:
+            try:
+                expanded_diag = diag.expand(self.shape[:-1])
+            except RuntimeError:
+                raise RuntimeError(
+                    "add_diag for LazyTensor of size {} received invalid diagonal of size {}.".format(
+                        self.shape, diag_shape
+                    )
                 )
-            )
+            diag_tensor = DiagLazyTensor(expanded_diag)
 
-        return AddedDiagLazyTensor(self, DiagLazyTensor(expanded_diag))
+        return AddedDiagLazyTensor(self, diag_tensor)
 
     def add_jitter(self, jitter_val=1e-3):
         """
@@ -736,10 +747,10 @@ class LazyTensor(ABC):
         Returns:
             (LazyTensor) Cholesky factor (lower triangular)
         """
-        res = self._cholesky()
+        chol = self._cholesky(upper=False)
         if upper:
-            res = res.transpose(-1, -2)
-        return res
+            chol = chol._transpose_nonbatch()
+        return chol
 
     def clone(self):
         """
@@ -839,6 +850,24 @@ class LazyTensor(ABC):
         """
         return self.ndimension()
 
+    def double(self, device_id=None):
+        """
+        This method operates identically to :func:`torch.Tensor.double`.
+        """
+        new_args = []
+        new_kwargs = {}
+        for arg in self._args:
+            if hasattr(arg, "double"):
+                new_args.append(arg.double())
+            else:
+                new_args.append(arg)
+        for name, val in self._kwargs.items():
+            if hasattr(val, "double"):
+                new_kwargs[name] = val.double()
+            else:
+                new_kwargs[name] = val
+        return self.__class__(*new_args, **new_kwargs)
+
     @property
     def dtype(self):
         return self._args[0].dtype
@@ -955,26 +984,16 @@ class LazyTensor(ABC):
                 "Got a {} of size {}.".format(self.__class__.__name__, self.size())
             )
 
-        if self.dim() == 2 and tensor.dim() == 1:
-            if self.shape[-1] != tensor.numel():
-                raise RuntimeError(
-                    "LazyTensor (size={}) cannot be multiplied with right-hand-side Tensor (size={}).".format(
-                        self.shape, tensor.shape
-                    )
-                )
-        elif self.dim() != tensor.dim():
-            raise RuntimeError(
-                "LazyTensor (size={}) and right-hand-side Tensor (size={}) should have the same number "
-                "of dimensions.".format(self.shape, tensor.shape)
-            )
-        elif self.shape[-1] != tensor.shape[-2]:
+        try:
+            result_shape = _matmul_broadcast_shape(self.shape, tensor.shape)
+        except RuntimeError:
             raise RuntimeError(
                 "LazyTensor (size={}) cannot be multiplied with right-hand-side Tensor (size={}).".format(
                     self.shape, tensor.shape
                 )
             )
 
-        args = (tensor,) + self.representation()
+        args = (tensor.expand(*result_shape[:-2], *tensor.shape[-2:]),) + self.representation()
         func = InvQuad.apply
         inv_quad_term = func(self.representation_tree(), *args)
 
@@ -998,8 +1017,9 @@ class LazyTensor(ABC):
         # Special case: use Cholesky to compute these terms
         if settings.fast_computations.log_prob.off() or (self.size(-1) <= settings.max_cholesky_size.value()):
             from .chol_lazy_tensor import CholLazyTensor
+            from .triangular_lazy_tensor import TriangularLazyTensor
 
-            cholesky = CholLazyTensor(self.cholesky())
+            cholesky = CholLazyTensor(TriangularLazyTensor(self.cholesky()))
             return cholesky.inv_quad_logdet(inv_quad_rhs=inv_quad_rhs, logdet=logdet, reduce_inv_quad=reduce_inv_quad)
 
         # Default: use modified batch conjugate gradients to compute these terms
@@ -1249,6 +1269,7 @@ class LazyTensor(ABC):
                 "Invalid repeat arguments {}. Currently, repeat only works to create repeated "
                 "batches of a 2D LazyTensor.".format(tuple(sizes))
             )
+
         return BatchRepeatLazyTensor(self, batch_repeat=torch.Size(sizes[:-2]))
 
     def representation(self):
@@ -1301,7 +1322,7 @@ class LazyTensor(ABC):
         return self
 
     @cached(name="root_decomposition")
-    def root_decomposition(self):
+    def root_decomposition(self, method: Optional[str] = None):
         """
         Returns a (usually low-rank) root decomposition lazy tensor of a PSD matrix.
         This can be used for sampling from a Gaussian distribution, or for obtaining a
@@ -1316,22 +1337,45 @@ class LazyTensor(ABC):
                 "Got a {} of size {}.".format(self.__class__.__name__, self.size())
             )
 
-        if (
-            self.size(-1) <= settings.max_cholesky_size.value()
-            or settings.fast_computations.covar_root_decomposition.off()
-        ):
+        if method is None:
+            if (
+                self.size(-1) <= settings.max_cholesky_size.value()
+                or settings.fast_computations.covar_root_decomposition.off()
+            ):
+                method = "cholesky"
+            else:
+                method = "lanczos"
+
+        if method == "cholesky":
             try:
                 res = self.cholesky()
                 return CholLazyTensor(res)
-
             except RuntimeError as e:
                 warnings.warn(
-                    "Runtime Error when computing Cholesky decomposition: {}. Using RootDecomposition.".format(e),
+                    f"Runtime Error when computing Cholesky decomposition: {e}. Using RootDecomposition.".format(e),
                     NumericalWarning,
                 )
+                method = "symeig"
 
-        res = self._root_decomposition()
-        return RootLazyTensor(res)
+        if method == "pivoted_cholesky":
+            return RootLazyTensor(pivoted_cholesky(self.evaluate(), max_iter=self._root_decomposition_size()))
+
+        if method == "symeig":
+            evals, evecs = self.symeig(eigenvectors=True)
+            # TODO: only use non-zero evals (req. dealing w/ batches...)
+            F = evecs * evals.clamp(0.0).sqrt().unsqueeze(-2)
+            return RootLazyTensor(F)
+
+        if method == "svd":
+            U, S, _ = self.svd()
+            # TODO: only use non-zero singular values (req. dealing w/ batches...)
+            F = U * S.sqrt().unsqueeze(-2)
+            return RootLazyTensor(F)
+
+        if method == "lanczos":
+            return RootLazyTensor(self._root_decomposition())
+
+        raise RuntimeError(f"Unknown method '{method}'")
 
     @cached(name="root_inv_decomposition")
     def root_inv_decomposition(self, initial_vectors=None, test_vectors=None):
@@ -1446,6 +1490,26 @@ class LazyTensor(ABC):
     def shape(self):
         return self.size()
 
+    def sqrt_inv_matmul(self, rhs, lhs=None):
+        """
+        If A is positive definite, computes either lhs A^{-1/2} rhs or A^{-1/2} rhs.
+        """
+        squeeze = False
+        if rhs.dim() == 1:
+            rhs = rhs.unsqueeze(-1)
+            squeeze = True
+
+        func = SqrtInvMatmul()
+        sqrt_inv_matmul_res, inv_quad_res = func.apply(self.representation_tree(), rhs, lhs, *self.representation())
+
+        if squeeze:
+            sqrt_inv_matmul_res = sqrt_inv_matmul_res.squeeze(-1)
+
+        if lhs is None:
+            return sqrt_inv_matmul_res
+        else:
+            return sqrt_inv_matmul_res, inv_quad_res
+
     def sum(self, dim=None):
         """
         Sum the LazyTensor across a dimension.
@@ -1492,6 +1556,45 @@ class LazyTensor(ABC):
         else:
             raise ValueError("Invalid dim ({}) for LazyTensor of size {}".format(orig_dim, self.shape))
 
+    def svd(self) -> Tuple["LazyTensor", Tensor, "LazyTensor"]:
+        """
+        Compute the SVD of the lazy tensor `M` s.t. `M = U @ S @ V.T`.
+        This can be very slow for large tensors. Should be special-cased for tensors with particular structure.
+        Does NOT sort the sigular values.
+
+        Returns:
+            :obj:`~gpytorch.lazy.LazyTensor`:
+                The left singular vectors (`U`).
+            :obj:`torch.Tensor`:
+                The singular values (`S`).
+            :obj:`~gpytorch.lazy.LazyTensor`:
+                The right singular vectors (`V`).
+        """
+        return self._svd()
+
+    @cached(name="symeig")
+    def symeig(self, eigenvectors: bool = False) -> Tuple[Tensor, Optional["LazyTensor"]]:
+        """
+        Compute the symmetric eigendecomposition of the lazy tensor. This can be very
+        slow for large tensors. Should be special-cased for tensors with particular
+        structure. Does NOT sort the eigenvalues.
+
+        Args:
+            :attr:`eigenvectors` (bool): If True, compute the eigenvectors in addition to the eigenvalues.
+        Returns:
+            :obj:`torch.Tensor`:
+                The eigenvalues.
+            :obj:`~gpytorch.lazy.LazyTensor`:
+                The eigenvectors. If `eigenvectors=False`, this is None. Otherwise, this LazyTensor
+                contains the orthonormal eigenvectors of the matrix.
+        """
+        try:
+            evals, evecs = pop_from_cache(self, "symeig", eigenvectors=True)
+            return evals, None
+        except CachingError:
+            pass
+        return self._symeig(eigenvectors=eigenvectors)
+
     def to(self, device_id):
         """
         A device-agnostic method of moving the lazy_tensor to the specified device.
@@ -1585,15 +1688,33 @@ class LazyTensor(ABC):
             :obj:`torch.tensor`:
                 Samples from MVN (num_samples x batch_size x num_dim) or (num_samples x num_dim)
         """
-        if self.size()[-2:] == torch.Size([1, 1]):
-            covar_root = self.evaluate().sqrt()
-        else:
-            covar_root = self.root_decomposition().root
+        from ..utils.contour_integral_quad import contour_integral_quad
 
-        base_samples = torch.randn(
-            *self.batch_shape, covar_root.size(-1), num_samples, dtype=self.dtype, device=self.device
-        )
-        samples = covar_root.matmul(base_samples).permute(-1, *range(self.dim() - 1)).contiguous()
+        if settings.ciq_samples.on():
+            base_samples = torch.randn(
+                *self.batch_shape, self.size(-1), num_samples, dtype=self.dtype, device=self.device
+            )
+            base_samples = base_samples.permute(-1, *range(self.dim() - 1)).contiguous()
+            base_samples = base_samples.unsqueeze(-1)
+            solves, weights, _, _ = contour_integral_quad(
+                self.evaluate_kernel(),
+                base_samples,
+                inverse=False,
+                num_contour_quadrature=settings.num_contour_quadrature.value(),
+            )
+
+            return (solves * weights).sum(0).squeeze(-1)
+
+        else:
+            if self.size()[-2:] == torch.Size([1, 1]):
+                covar_root = self.evaluate().sqrt()
+            else:
+                covar_root = self.root_decomposition().root
+
+            base_samples = torch.randn(
+                *self.batch_shape, covar_root.size(-1), num_samples, dtype=self.dtype, device=self.device
+            )
+            samples = covar_root.matmul(base_samples).permute(-1, *range(self.dim() - 1)).contiguous()
 
         return samples
 
@@ -1624,7 +1745,9 @@ class LazyTensor(ABC):
         elif isinstance(other, Tensor):
             other = lazify(other)
             shape = _mul_broadcast_shape(self.shape, other.shape)
-            return SumLazyTensor(self.expand(shape), other.expand(shape))
+            new_self = self if self.shape[:-2] == shape[:-2] else self._expand_batch(shape[:-2])
+            new_other = other if other.shape[:-2] == shape[:-2] else other._expand_batch(shape[:-2])
+            return SumLazyTensor(new_self, new_other)
         else:
             return SumLazyTensor(self, other)
 
@@ -1735,6 +1858,33 @@ class LazyTensor(ABC):
         # We're done!
         return res
 
+    @cached(name="svd")
+    def _svd(self) -> Tuple["LazyTensor", Tensor, "LazyTensor"]:
+        """Method that allows implementing special-cased SVD computation. Should not be called directly"""
+        # Using symeig is preferable here for psd LazyTensors.
+        # Will need to overwrite this function for non-psd LazyTensors.
+        evals, evecs = self.symeig(eigenvectors=True)
+        signs = torch.sign(evals)
+        U = evecs * signs.unsqueeze(-2)
+        S = torch.abs(evals)
+        V = evecs
+        return U, S, V
+
+    def _symeig(self, eigenvectors: bool = False) -> Tuple[Tensor, Optional["LazyTensor"]]:
+        """Method that allows implementing special-cased symeig computation. Should not be called directly"""
+        from gpytorch.lazy.non_lazy_tensor import NonLazyTensor
+
+        dtype = self.dtype  # perform decomposition in double precision for numerical stability
+        # TODO: Use fp64 registry once #1213 is addressed
+        evals, evecs = torch.symeig(self.evaluate().to(dtype=torch.double), eigenvectors=eigenvectors)
+        # chop any negative eigenvalues. TODO: warn if evals are significantly negative
+        evals = evals.clamp_min(0.0).to(dtype=dtype)
+        if eigenvectors:
+            evecs = NonLazyTensor(evecs.to(dtype=dtype))
+        else:
+            evecs = None
+        return evals, evecs
+
     def __matmul__(self, other):
         return self.matmul(other)
 
