@@ -8,6 +8,7 @@ import torch
 from .. import settings
 from ..lazy import (
     BatchRepeatLazyTensor,
+    ConstantMulLazyTensor,
     InterpolatedLazyTensor,
     LazyEvaluatedKernelTensor,
     MatmulLazyTensor,
@@ -20,11 +21,7 @@ from ..lazy import (
 )
 from ..utils.cholesky import psd_safe_cholesky
 from ..utils.interpolation import left_interp, left_t_interp
-from ..utils.memoize import add_to_cache, cached
-
-
-def clear_cache_hook(module, *args, **kwargs):
-    module._memoize_cache = {}
+from ..utils.memoize import add_to_cache, cached, clear_cache_hook, pop_from_cache
 
 
 def prediction_strategy(train_inputs, train_prior_dist, train_labels, likelihood):
@@ -171,7 +168,8 @@ class DefaultPredictionStrategy(object):
         small_system_rhs = targets - fant_mean - ftcm
         small_system_rhs = small_system_rhs.unsqueeze(-1)
         # Schur complement of a spd matrix is guaranteed to be positive definite
-        fant_cache_lower = torch.cholesky_solve(small_system_rhs, psd_safe_cholesky(schur_complement))
+        schur_cholesky = psd_safe_cholesky(schur_complement, jitter=settings.cholesky_jitter.value())
+        fant_cache_lower = torch.cholesky_solve(small_system_rhs, schur_cholesky)
 
         # Get "a", the new upper portion of the cache corresponding to the old training points.
         fant_cache_upper = self.mean_cache.unsqueeze(-1) - fant_solve.matmul(fant_cache_lower)
@@ -202,24 +200,41 @@ class DefaultPredictionStrategy(object):
         batch_shape = fant_train_covar.shape[:-2]
 
         L_inverse = self.covar_cache
-        L = delazify(self.lik_train_train_covar.root_decomposition().root)
+        L = self.lik_train_train_covar.root_decomposition().root
         m, n = L.shape[-2:]
 
         lower_left = fant_train_covar.matmul(L_inverse)
-        schur_root = psd_safe_cholesky(fant_fant_covar - lower_left.matmul(lower_left.transpose(-2, -1)))
+        schur = fant_fant_covar - lower_left.matmul(lower_left.transpose(-2, -1))
+        schur_root = psd_safe_cholesky(schur, jitter=settings.cholesky_jitter.value())
 
         # Form new root Z = [L 0; lower_left schur_root]
+
+        # # TODO: Special case triangular case once #1102 goes in
+        # if isinstance(L, TriangularLazyTensor):
+        #     # The whole thing is triangular, we can just do two triangular solves
+        #     ...
+        # else:
+
+        L = delazify(L)
         num_fant = schur_root.size(-2)
-        m, n = L.shape[-2:]
         new_root = torch.zeros(*batch_shape, m + num_fant, n + num_fant, device=L.device, dtype=L.dtype)
         new_root[..., :m, :n] = L
         new_root[..., m:, :n] = lower_left
         new_root[..., m:, n:] = schur_root
 
         # Use pseudo-inverse of Z as new inv root
-        Q, R = torch.qr(new_root)
+
+        if new_root.shape[-1] <= 2048:
+            # Dispatch to CPU so long as pytorch/pytorch#22573 is not fixed
+            device = new_root.device
+            Q, R = torch.qr(new_root.cpu())
+            Q = Q.to(device)
+            R = R.to(device)
+        else:
+            Q, R = torch.qr(new_root)
+
         Rdiag = torch.diagonal(R, dim1=-2, dim2=-1)
-        # if R is almost singular, add jitter (Rdiag is a view, so this works)
+        # if R is almost singular, add jitter
         zeroish = Rdiag.abs() < 1e-6
         if torch.any(zeroish):
             # can't use in-place operation here b/c it would mess up backward pass
@@ -248,8 +263,8 @@ class DefaultPredictionStrategy(object):
             root=new_root,
             inv_root=new_covar_cache,
         )
-        fant_strat._memoize_cache = {"mean_cache": fant_mean_cache, "covar_cache": new_covar_cache}
-
+        add_to_cache(fant_strat, "mean_cache", fant_mean_cache)
+        add_to_cache(fant_strat, "covar_cache", new_covar_cache)
         return fant_strat
 
     @property
@@ -529,10 +544,9 @@ class InterpolatedPredictionStrategy(DefaultPredictionStrategy):
         test_interp_values = test_train_covar.left_interp_values
 
         precomputed_cache = self.covar_cache
-        if (settings.fast_pred_samples.on() and precomputed_cache[0] is None) or (
-            settings.fast_pred_samples.off() and precomputed_cache[1] is None
-        ):
-            self._memoize_cache.pop("covar_cache")
+        fps = settings.fast_pred_samples.on()
+        if (fps and precomputed_cache[0] is None) or (not fps and precomputed_cache[1] is None):
+            pop_from_cache(self, "covar_cache")
             precomputed_cache = self.covar_cache
 
         # Compute the exact predictive posterior
@@ -561,7 +575,7 @@ class SumPredictionStrategy(DefaultPredictionStrategy):
         return sub_strategies
 
     def _exact_predictive_covar_inv_quad_form_cache(self, train_train_covar_inv_root, test_train_covar):
-        test_train_covar = test_train_covar.evaluate_kernel()
+        test_train_covar = lazify(test_train_covar).evaluate_kernel()
         if not isinstance(test_train_covar, SumLazyTensor):
             return super(SumPredictionStrategy, self)._exact_predictive_covar_inv_quad_form_cache(
                 train_train_covar_inv_root, test_train_covar
@@ -575,7 +589,7 @@ class SumPredictionStrategy(DefaultPredictionStrategy):
     def _exact_predictive_covar_inv_quad_form_root(self, precomputed_cache, test_train_covar):
         # Here the precomputed cache is a list
         # where each component in the list is the precomputed cache for each component lazy tensor
-        test_train_covar = test_train_covar.evaluate_kernel()
+        test_train_covar = lazify(test_train_covar).evaluate_kernel()
         if not isinstance(test_train_covar, SumLazyTensor):
             return super(SumPredictionStrategy, self)._exact_predictive_covar_inv_quad_form_root(
                 precomputed_cache, test_train_covar
@@ -587,3 +601,58 @@ class SumPredictionStrategy(DefaultPredictionStrategy):
                     self._sub_strategies, precomputed_cache, test_train_covar.evaluate_kernel().lazy_tensors
                 )
             )
+
+
+class RFFPredictionStrategy(DefaultPredictionStrategy):
+    def __init__(self, train_inputs, train_prior_dist, train_labels, likelihood):
+        super().__init__(train_inputs, train_prior_dist, train_labels, likelihood)
+        self.train_prior_dist = self.train_prior_dist.__class__(
+            self.train_prior_dist.mean, self.train_prior_dist.lazy_covariance_matrix.evaluate_kernel()
+        )
+
+    def get_fantasy_strategy(self, inputs, targets, full_inputs, full_targets, full_output, **kwargs):
+        raise NotImplementedError("Fantasy observation updates not yet supported for models using RFFs")
+
+    @property
+    @cached(name="covar_cache")
+    def covar_cache(self):
+        lt = self.train_prior_dist.lazy_covariance_matrix
+        if isinstance(lt, ConstantMulLazyTensor):
+            constant = lt.expanded_constant
+            lt = lt.base_lazy_tensor
+        else:
+            constant = torch.tensor(1.0, dtype=lt.dtype, device=lt.device)
+
+        train_factor = lt.root.evaluate()
+        train_train_covar = self.lik_train_train_covar
+        inner_term = (
+            torch.eye(train_factor.size(-1), dtype=train_factor.dtype, device=train_factor.device)
+            - (train_factor.transpose(-1, -2) @ train_train_covar.inv_matmul(train_factor)) * constant
+        )
+        return psd_safe_cholesky(inner_term)
+
+    def exact_prediction(self, joint_mean, joint_covar):
+        # Find the components of the distribution that contain test data
+        test_mean = joint_mean[..., self.num_train :]
+        test_test_covar = joint_covar[..., self.num_train :, self.num_train :].evaluate_kernel()
+        test_train_covar = joint_covar[..., self.num_train :, : self.num_train].evaluate_kernel()
+
+        return (
+            self.exact_predictive_mean(test_mean, test_train_covar),
+            self.exact_predictive_covar(test_test_covar, test_train_covar),
+        )
+
+    def exact_predictive_covar(self, test_test_covar, test_train_covar):
+        if settings.skip_posterior_variances.on():
+            return ZeroLazyTensor(*test_test_covar.size())
+
+        if isinstance(test_test_covar, ConstantMulLazyTensor):
+            constant = test_test_covar.expanded_constant
+            test_test_covar = test_test_covar.base_lazy_tensor
+        else:
+            constant = torch.tensor(1.0, dtype=test_test_covar.dtype, device=test_test_covar.device)
+
+        covar_cache = self.covar_cache
+        factor = test_test_covar.root.evaluate() * constant.sqrt()
+        res = RootLazyTensor(factor @ covar_cache)
+        return res
