@@ -1,5 +1,7 @@
 #!/usr/bin/env python3
 
+import torch
+
 
 class _feature_flag(object):
     _state = False
@@ -51,6 +53,49 @@ class _value_context(object):
         return False
 
 
+class _dtype_value_context(object):
+    _global_float_value = None
+    _global_double_value = None
+    _global_half_value = None
+
+    @classmethod
+    def value(cls, dtype):
+        if torch.is_tensor(dtype):
+            dtype = dtype.dtype
+        if dtype == torch.float:
+            return cls._global_float_value
+        elif dtype == torch.double:
+            return cls._global_double_value
+        elif dtype == torch.half:
+            return cls._global_half_value
+        else:
+            raise RuntimeError(f"Unsupported dtype for {cls.__name__}.")
+
+    @classmethod
+    def _set_value(cls, float_value, double_value, half_value):
+        if float_value is not None:
+            cls._global_float_value = float_value
+        if double_value is not None:
+            cls._global_double_value = double_value
+        if half_value is not None:
+            cls._global_half_value = half_value
+
+    def __init__(self, float=None, double=None, half=None):
+        self._orig_float_value = self.__class__.value()
+        self._instance_float_value = float
+        self._orig_double_value = self.__class__.value()
+        self._instance_double_value = double
+        self._orig_half_value = self.__class__.value()
+        self._instance_half_value = half
+
+    def __enter__(self,):
+        self.__class__._set_value(self._instance_float_value, self._instance_double_value, self._instance_half_value)
+
+    def __exit__(self, *args):
+        self.__class__._set_value(self._orig_float_value, self._orig_double_value, self._orig_half_value)
+        return False
+
+
 class _fast_covar_root_decomposition(_feature_flag):
     r"""
     This feature flag controls how matrix root decompositions (:math:`K = L L^\top`) are computed
@@ -338,6 +383,26 @@ class max_cg_iterations(_value_context):
     _global_value = 1000
 
 
+class min_variance(_dtype_value_context):
+    """
+    The minimum variance that can be returned from :obj:`~gpytorch.distributions.MultivariateNormal#variance`
+    (default 1e-6). If variances are smaller than this, they are rounded up and a warning is raised.
+    """
+
+    _global_float_value = 1e-6
+    _global_double_value = 1e-10
+    _global_half_value = 1e-3
+
+
+class cholesky_jitter(_value_context):
+    """
+    The jitter value passed to `psd_safe_cholesky` when using cholesky solves.
+    Default: None
+    """
+
+    _global_value = None
+
+
 class cg_tolerance(_value_context):
     """
     Relative residual tolerance to use for terminating CG.
@@ -348,6 +413,14 @@ class cg_tolerance(_value_context):
     _global_value = 1
 
 
+class ciq_samples(_feature_flag):
+    """
+    Whether to draw samples using CIQ or not
+    """
+
+    _state = False
+
+
 class preconditioner_tolerance(_value_context):
     """
     Diagonal trace tolerance to use for checking preconditioner convergence.
@@ -398,7 +471,7 @@ class max_preconditioner_size(_value_context):
     """
     The maximum size of preconditioner to use. 0 corresponds to turning
     preconditioning off. When enabled, usually a value of around ~10 works fairly well.
-    Default: 0
+    Default: 15
     """
 
     _global_value = 15
@@ -435,6 +508,25 @@ class min_preconditioning_size(_value_context):
     _global_value = 2000
 
 
+class minres_tolerance(_value_context):
+    """
+    Relative update term tolerance to use for terminating MINRES.
+
+    Default: 1e-4
+    """
+
+    _global_value = 1e-4
+
+
+class num_contour_quadrature(_value_context):
+    """
+    The number of quadrature points to compute CIQ.
+    Default: 15
+    """
+
+    _global_value = 15
+
+
 class num_likelihood_samples(_value_context):
     """
     The number of samples to draw from a latent GP when computing a likelihood
