@@ -1,5 +1,6 @@
 #!/usr/bin/env python3
 
+import itertools
 import math
 from abc import abstractmethod
 from itertools import combinations, product
@@ -42,9 +43,22 @@ class RectangularLazyTensorTestCase(BaseTestCase):
         res.backward(gradient=grad)
         actual.backward(gradient=grad)
         for arg, arg_copy in zip(lazy_tensor.representation(), lazy_tensor_copy.representation()):
-            if arg_copy.grad is not None:
+            if arg_copy.requires_grad and arg_copy.is_leaf and arg_copy.grad is not None:
                 self.assertAllClose(arg.grad, arg_copy.grad, rtol=1e-3)
 
+    def test_add(self):
+        lazy_tensor = self.create_lazy_tensor()
+        evaluated = self.evaluate_lazy_tensor(lazy_tensor)
+
+        rhs = torch.randn(lazy_tensor.shape)
+        self.assertAllClose((lazy_tensor + rhs).evaluate(), evaluated + rhs)
+
+        rhs = torch.randn(lazy_tensor.matrix_shape)
+        self.assertAllClose((lazy_tensor + rhs).evaluate(), evaluated + rhs)
+
+        rhs = torch.randn(2, *lazy_tensor.shape)
+        self.assertAllClose((lazy_tensor + rhs).evaluate(), evaluated + rhs)
+
     def test_matmul_vec(self):
         lazy_tensor = self.create_lazy_tensor()
         rhs = torch.randn(lazy_tensor.size(-1))
@@ -284,7 +298,7 @@ class LazyTensorTestCase(RectangularLazyTensorTestCase):
                 res.backward(gradient=grad)
                 actual.backward(gradient=grad)
                 for arg, arg_copy in zip(lazy_tensor.representation(), lazy_tensor_copy.representation()):
-                    if arg_copy.grad is not None:
+                    if arg_copy.requires_grad and arg_copy.is_leaf and arg_copy.grad is not None:
                         self.assertAllClose(arg.grad, arg_copy.grad, rtol=0.03, atol=1e-5)
                 self.assertAllClose(rhs.grad, rhs_copy.grad, rtol=0.03, atol=1e-5)
                 if lhs is not None:
@@ -363,6 +377,15 @@ class LazyTensorTestCase(RectangularLazyTensorTestCase):
                 actual[..., i, i] = actual[..., i, i] + other_diag[..., i]
             self.assertAllClose(res, actual, rtol=1e-2, atol=1e-5)
 
+    def test_cholesky(self):
+        lazy_tensor = self.create_lazy_tensor()
+        evaluated = self.evaluate_lazy_tensor(lazy_tensor)
+        for upper in (False, True):
+            res = lazy_tensor.cholesky(upper=upper).evaluate()
+            actual = torch.cholesky(evaluated, upper=upper)
+            self.assertAllClose(res, actual, rtol=1e-3, atol=1e-5)
+            # TODO: Check gradients
+
     def test_diag(self):
         lazy_tensor = self.create_lazy_tensor()
         evaluated = self.evaluate_lazy_tensor(lazy_tensor)
@@ -491,3 +514,173 @@ class LazyTensorTestCase(RectangularLazyTensorTestCase):
             samples = lazy_tensor.zero_mean_mvn_samples(50000)
             sample_covar = samples.unsqueeze(-1).matmul(samples.unsqueeze(-2)).mean(0)
             self.assertAllClose(sample_covar, evaluated, rtol=0.3, atol=0.3)
+
+    def test_sqrt_inv_matmul(self):
+        lazy_tensor = self.create_lazy_tensor().requires_grad_(True)
+        if len(lazy_tensor.batch_shape):
+            return
+
+        lazy_tensor_copy = lazy_tensor.clone().detach_().requires_grad_(True)
+        evaluated = self.evaluate_lazy_tensor(lazy_tensor_copy)
+        evaluated.register_hook(_ensure_symmetric_grad)
+
+        # Create a test right hand side and left hand side
+        rhs = torch.randn(*lazy_tensor.shape[:-1], 3).requires_grad_(True)
+        lhs = torch.randn(*lazy_tensor.shape[:-2], 2, lazy_tensor.size(-1)).requires_grad_(True)
+        rhs_copy = rhs.clone().detach().requires_grad_(True)
+        lhs_copy = lhs.clone().detach().requires_grad_(True)
+
+        # Perform forward pass
+        with gpytorch.settings.max_cg_iterations(200):
+            sqrt_inv_matmul_res, inv_quad_res = lazy_tensor.sqrt_inv_matmul(rhs, lhs)
+        evals, evecs = evaluated.symeig(eigenvectors=True)
+        matrix_inv_root = evecs @ (evals.sqrt().reciprocal().unsqueeze(-1) * evecs.transpose(-1, -2))
+        sqrt_inv_matmul_actual = lhs_copy @ matrix_inv_root @ rhs_copy
+        inv_quad_actual = (lhs_copy @ matrix_inv_root).pow(2).sum(dim=-1)
+
+        # Check forward pass
+        self.assertAllClose(sqrt_inv_matmul_res, sqrt_inv_matmul_actual, rtol=1e-4, atol=1e-3)
+        self.assertAllClose(inv_quad_res, inv_quad_actual, rtol=1e-4, atol=1e-3)
+
+        # Perform backward pass
+        sqrt_inv_matmul_grad = torch.randn_like(sqrt_inv_matmul_res)
+        inv_quad_grad = torch.randn_like(inv_quad_res)
+        ((sqrt_inv_matmul_res * sqrt_inv_matmul_grad).sum() + (inv_quad_res * inv_quad_grad).sum()).backward()
+        ((sqrt_inv_matmul_actual * sqrt_inv_matmul_grad).sum() + (inv_quad_actual * inv_quad_grad).sum()).backward()
+
+        # Check grads
+        self.assertAllClose(rhs.grad, rhs_copy.grad, rtol=1e-4, atol=1e-3)
+        self.assertAllClose(lhs.grad, lhs_copy.grad, rtol=1e-4, atol=1e-3)
+        for arg, arg_copy in zip(lazy_tensor.representation(), lazy_tensor_copy.representation()):
+            if arg_copy.requires_grad and arg_copy.is_leaf and arg_copy.grad is not None:
+                self.assertAllClose(arg.grad, arg_copy.grad, rtol=1e-4, atol=1e-3)
+
+    def test_sqrt_inv_matmul_no_lhs(self):
+        lazy_tensor = self.create_lazy_tensor().requires_grad_(True)
+        if len(lazy_tensor.batch_shape):
+            return
+
+        lazy_tensor_copy = lazy_tensor.clone().detach_().requires_grad_(True)
+        evaluated = self.evaluate_lazy_tensor(lazy_tensor_copy)
+        evaluated.register_hook(_ensure_symmetric_grad)
+
+        # Create a test right hand side and left hand side
+        rhs = torch.randn(*lazy_tensor.shape[:-1], 3).requires_grad_(True)
+        rhs_copy = rhs.clone().detach().requires_grad_(True)
+
+        # Perform forward pass
+        with gpytorch.settings.max_cg_iterations(200):
+            sqrt_inv_matmul_res = lazy_tensor.sqrt_inv_matmul(rhs)
+        evals, evecs = evaluated.symeig(eigenvectors=True)
+        matrix_inv_root = evecs @ (evals.sqrt().reciprocal().unsqueeze(-1) * evecs.transpose(-1, -2))
+        sqrt_inv_matmul_actual = matrix_inv_root @ rhs_copy
+
+        # Check forward pass
+        self.assertAllClose(sqrt_inv_matmul_res, sqrt_inv_matmul_actual, rtol=1e-4, atol=1e-3)
+
+        # Perform backward pass
+        sqrt_inv_matmul_grad = torch.randn_like(sqrt_inv_matmul_res)
+        ((sqrt_inv_matmul_res * sqrt_inv_matmul_grad).sum()).backward()
+        ((sqrt_inv_matmul_actual * sqrt_inv_matmul_grad).sum()).backward()
+
+        # Check grads
+        self.assertAllClose(rhs.grad, rhs_copy.grad, rtol=1e-4, atol=1e-3)
+        for arg, arg_copy in zip(lazy_tensor.representation(), lazy_tensor_copy.representation()):
+            if arg_copy.requires_grad and arg_copy.is_leaf and arg_copy.grad is not None:
+                self.assertAllClose(arg.grad, arg_copy.grad, rtol=1e-4, atol=1e-3)
+
+    def test_symeig(self):
+        lazy_tensor = self.create_lazy_tensor().requires_grad_(True)
+        lazy_tensor_copy = lazy_tensor.clone().detach_().requires_grad_(True)
+        evaluated = self.evaluate_lazy_tensor(lazy_tensor_copy)
+
+        # Perform forward pass
+        evals_unsorted, evecs_unsorted = lazy_tensor.symeig(eigenvectors=True)
+        evecs_unsorted = evecs_unsorted.evaluate()
+
+        # since LazyTensor.symeig does not sort evals, we do this here for the check
+        evals, idxr = torch.sort(evals_unsorted, dim=-1, descending=False)
+        evecs = torch.gather(evecs_unsorted, dim=-1, index=idxr.unsqueeze(-2).expand(evecs_unsorted.shape))
+
+        evals_actual, evecs_actual = torch.symeig(evaluated.double(), eigenvectors=True)
+        evals_actual = evals_actual.to(dtype=evaluated.dtype)
+        evecs_actual = evecs_actual.to(dtype=evaluated.dtype)
+
+        # Check forward pass
+        self.assertAllClose(evals, evals_actual, rtol=1e-4, atol=1e-3)
+        lt_from_eigendecomp = evecs @ torch.diag_embed(evals) @ evecs.transpose(-1, -2)
+        self.assertAllClose(lt_from_eigendecomp, evaluated, rtol=1e-4, atol=1e-3)
+
+        # if there are repeated evals, we'll skip checking the eigenvectors for those
+        any_evals_repeated = False
+        evecs_abs, evecs_actual_abs = evecs.abs(), evecs_actual.abs()
+        for idx in itertools.product(*[range(b) for b in evals_actual.shape[:-1]]):
+            eval_i = evals_actual[idx]
+            if torch.unique(eval_i.detach()).shape[-1] == eval_i.shape[-1]:  # detach to avoid pytorch/pytorch#41389
+                self.assertAllClose(evecs_abs[idx], evecs_actual_abs[idx], rtol=1e-4, atol=1e-3)
+            else:
+                any_evals_repeated = True
+
+        # Perform backward pass
+        symeig_grad = torch.randn_like(evals)
+        ((evals * symeig_grad).sum()).backward()
+        ((evals_actual * symeig_grad).sum()).backward()
+
+        # Check grads if there were no repeated evals
+        if not any_evals_repeated:
+            for arg, arg_copy in zip(lazy_tensor.representation(), lazy_tensor_copy.representation()):
+                if arg_copy.requires_grad and arg_copy.is_leaf and arg_copy.grad is not None:
+                    self.assertAllClose(arg.grad, arg_copy.grad, rtol=1e-4, atol=1e-3)
+
+        # Test with eigenvectors=False
+        _, evecs = lazy_tensor.symeig(eigenvectors=False)
+        self.assertIsNone(evecs)
+
+    def test_svd(self):
+        lazy_tensor = self.create_lazy_tensor().requires_grad_(True)
+        lazy_tensor_copy = lazy_tensor.clone().detach_().requires_grad_(True)
+        evaluated = self.evaluate_lazy_tensor(lazy_tensor_copy)
+
+        # Perform forward pass
+        U_unsorted, S_unsorted, V_unsorted = lazy_tensor.svd()
+        U_unsorted, V_unsorted = U_unsorted.evaluate(), V_unsorted.evaluate()
+
+        # since LazyTensor.svd does not sort the singular values, we do this here for the check
+        S, idxr = torch.sort(S_unsorted, dim=-1, descending=True)
+        idxr = idxr.unsqueeze(-2).expand(U_unsorted.shape)
+        U = torch.gather(U_unsorted, dim=-1, index=idxr)
+        V = torch.gather(V_unsorted, dim=-1, index=idxr)
+
+        # compute expected result from full tensor
+        U_actual, S_actual, V_actual = torch.svd(evaluated.double())
+        U_actual = U_actual.to(dtype=evaluated.dtype)
+        S_actual = S_actual.to(dtype=evaluated.dtype)
+        V_actual = V_actual.to(dtype=evaluated.dtype)
+
+        # Check forward pass
+        self.assertAllClose(S, S_actual, rtol=1e-4, atol=1e-3)
+        lt_from_svd = U @ torch.diag_embed(S) @ V.transpose(-1, -2)
+        self.assertAllClose(lt_from_svd, evaluated, rtol=1e-4, atol=1e-3)
+
+        # if there are repeated singular values, we'll skip checking the singular vectors
+        U_abs, U_actual_abs = U.abs(), U_actual.abs()
+        V_abs, V_actual_abs = V.abs(), V_actual.abs()
+        any_svals_repeated = False
+        for idx in itertools.product(*[range(b) for b in S_actual.shape[:-1]]):
+            Si = S_actual[idx]
+            if torch.unique(Si.detach()).shape[-1] == Si.shape[-1]:  # detach to avoid pytorch/pytorch#41389
+                self.assertAllClose(U_abs[idx], U_actual_abs[idx], rtol=1e-4, atol=1e-3)
+                self.assertAllClose(V_abs[idx], V_actual_abs[idx], rtol=1e-4, atol=1e-3)
+            else:
+                any_svals_repeated = True
+
+        # Perform backward pass
+        svd_grad = torch.randn_like(S)
+        ((S * svd_grad).sum()).backward()
+        ((S_actual * svd_grad).sum()).backward()
+
+        # Check grads if there were no repeated singular values
+        if not any_svals_repeated:
+            for arg, arg_copy in zip(lazy_tensor.representation(), lazy_tensor_copy.representation()):
+                if arg_copy.requires_grad and arg_copy.is_leaf and arg_copy.grad is not None:
+                    self.assertAllClose(arg.grad, arg_copy.grad, rtol=1e-4, atol=1e-3)
