@@ -3,12 +3,19 @@
 from ._variational_distribution import _VariationalDistribution
 from ._variational_strategy import _VariationalStrategy
 from .additive_grid_interpolation_variational_strategy import AdditiveGridInterpolationVariationalStrategy
+from .batch_decoupled_variational_strategy import BatchDecoupledVariationalStrategy
 from .cholesky_variational_distribution import CholeskyVariationalDistribution
 from .delta_variational_distribution import DeltaVariationalDistribution
 from .grid_interpolation_variational_strategy import GridInterpolationVariationalStrategy
+from .independent_multitask_variational_strategy import (
+    IndependentMultitaskVariationalStrategy,
+    MultitaskVariationalStrategy,
+)
+from .lmc_variational_strategy import LMCVariationalStrategy
 from .mean_field_variational_distribution import MeanFieldVariationalDistribution
-from .multitask_variational_strategy import MultitaskVariationalStrategy
+from .natural_variational_distribution import NaturalVariationalDistribution, _NaturalVariationalDistribution
 from .orthogonally_decoupled_variational_strategy import OrthogonallyDecoupledVariationalStrategy
+from .tril_natural_variational_distribution import TrilNaturalVariationalDistribution
 from .unwhitened_variational_strategy import UnwhitenedVariationalStrategy
 from .variational_strategy import VariationalStrategy
 from .whitened_variational_strategy import WhitenedVariationalStrategy
@@ -16,7 +23,10 @@ from .whitened_variational_strategy import WhitenedVariationalStrategy
 __all__ = [
     "_VariationalStrategy",
     "AdditiveGridInterpolationVariationalStrategy",
+    "BatchDecoupledVariationalStrategy",
     "GridInterpolationVariationalStrategy",
+    "IndependentMultitaskVariationalStrategy",
+    "LMCVariationalStrategy",
     "MultitaskVariationalStrategy",
     "OrthogonallyDecoupledVariationalStrategy",
     "VariationalStrategy",
@@ -26,4 +36,7 @@ __all__ = [
     "CholeskyVariationalDistribution",
     "MeanFieldVariationalDistribution",
     "DeltaVariationalDistribution",
+    "_NaturalVariationalDistribution",
+    "NaturalVariationalDistribution",
+    "TrilNaturalVariationalDistribution",
 ]
