@@ -1,6 +1,7 @@
 #!/usr/bin/env python3
 
 import unittest
+import warnings
 from math import exp, pi
 
 import gpytorch
@@ -9,7 +10,7 @@ from gpytorch.distributions import MultivariateNormal
 from gpytorch.kernels import RBFKernel, ScaleKernel
 from gpytorch.likelihoods import GaussianLikelihood
 from gpytorch.means import ConstantMean
-from gpytorch.priors import SmoothedBoxPrior, LogNormalPrior, UniformPrior
+from gpytorch.priors import SmoothedBoxPrior, UniformPrior
 from gpytorch.constraints import Positive
 from gpytorch.test.base_test_case import BaseTestCase
 from gpytorch.test.utils import least_used_cuda_device
@@ -100,6 +101,7 @@ class TestSimpleGPRegression(BaseTestCase, unittest.TestCase):
         )
 
     def test_posterior_latent_gp_and_likelihood_without_optimization(self, cuda=False):
+        warnings.simplefilter("ignore", gpytorch.utils.warnings.NumericalWarning)
         train_x, test_x, train_y, test_y = self._get_data(cuda=cuda)
         # We're manually going to set the hyperparameters to be ridiculous
         likelihood = GaussianLikelihood(noise_constraint=Positive())  # This test actually wants a noise < 1e-4
@@ -230,7 +232,7 @@ class TestSimpleGPRegression(BaseTestCase, unittest.TestCase):
         # Find optimal model hyperparameters
         gp_model.train()
         likelihood.train()
-        optimizer = optim.Adam(list(gp_model.parameters()) + list(likelihood.parameters()), lr=0.15)
+        optimizer = optim.Adam(gp_model.parameters(), lr=0.15)
         with gpytorch.beta_features.checkpoint_kernel(checkpoint), gpytorch.settings.fast_pred_var():
             for _ in range(20 if checkpoint else 50):
                 optimizer.zero_grad()
@@ -242,9 +244,6 @@ class TestSimpleGPRegression(BaseTestCase, unittest.TestCase):
             for param in gp_model.parameters():
                 self.assertTrue(param.grad is not None)
                 self.assertGreater(param.grad.norm().item(), 0)
-            for param in likelihood.parameters():
-                self.assertTrue(param.grad is not None)
-                self.assertGreater(param.grad.norm().item(), 0)
             optimizer.step()
 
         # Test the model
@@ -281,7 +280,7 @@ class TestSimpleGPRegression(BaseTestCase, unittest.TestCase):
         # Find optimal model hyperparameters
         gp_model.train()
         likelihood.train()
-        optimizer = optim.Adam(list(gp_model.parameters()) + list(likelihood.parameters()), lr=0.15)
+        optimizer = optim.Adam(gp_model.parameters(), lr=0.15)
         for _ in range(50):
             optimizer.zero_grad()
             with gpytorch.settings.debug(False):
@@ -293,9 +292,6 @@ class TestSimpleGPRegression(BaseTestCase, unittest.TestCase):
         for param in gp_model.parameters():
             self.assertTrue(param.grad is not None)
             self.assertGreater(param.grad.norm().item(), 0)
-        for param in likelihood.parameters():
-            self.assertTrue(param.grad is not None)
-            self.assertGreater(param.grad.norm().item(), 0)
         optimizer.step()
 
         train_x.requires_grad = True
@@ -350,7 +346,7 @@ class TestSimpleGPRegression(BaseTestCase, unittest.TestCase):
         # Find optimal model hyperparameters
         gp_model.train()
         likelihood.train()
-        optimizer = optim.Adam(list(gp_model.parameters()) + list(likelihood.parameters()), lr=0.15)
+        optimizer = optim.Adam(gp_model.parameters(), lr=0.15)
         for _ in range(50):
             optimizer.zero_grad()
             with gpytorch.settings.debug(False):
@@ -362,9 +358,6 @@ class TestSimpleGPRegression(BaseTestCase, unittest.TestCase):
         for param in gp_model.parameters():
             self.assertTrue(param.grad is not None)
             self.assertGreater(param.grad.norm().item(), 0)
-        for param in likelihood.parameters():
-            self.assertTrue(param.grad is not None)
-            self.assertGreater(param.grad.norm().item(), 0)
         optimizer.step()
 
         with gpytorch.settings.fast_pred_var():
@@ -421,7 +414,7 @@ class TestSimpleGPRegression(BaseTestCase, unittest.TestCase):
             # Find optimal model hyperparameters
             gp_model.train()
             likelihood.train()
-            optimizer = optim.Adam(list(gp_model.parameters()) + list(likelihood.parameters()), lr=0.1)
+            optimizer = optim.Adam(gp_model.parameters(), lr=0.1)
             for _ in range(50):
                 optimizer.zero_grad()
                 output = gp_model(train_x)
@@ -432,9 +425,6 @@ class TestSimpleGPRegression(BaseTestCase, unittest.TestCase):
             for param in gp_model.parameters():
                 self.assertTrue(param.grad is not None)
                 self.assertGreater(param.grad.norm().item(), 0)
-            for param in likelihood.parameters():
-                self.assertTrue(param.grad is not None)
-                self.assertGreater(param.grad.norm().item(), 0)
             optimizer.step()
 
             # Test the model
@@ -465,9 +455,9 @@ class TestSimpleGPRegression(BaseTestCase, unittest.TestCase):
 
         # Register normal GPyTorch priors
         gp_model.mean_module.register_prior("mean_prior", UniformPrior(-1, 1), "constant")
-        gp_model.covar_module.base_kernel.register_prior("lengthscale_prior", UniformPrior(0.01, 0.2), "lengthscale")
+        gp_model.covar_module.base_kernel.register_prior("lengthscale_prior", UniformPrior(0.01, 0.5), "lengthscale")
         gp_model.covar_module.register_prior("outputscale_prior", UniformPrior(1, 2), "outputscale")
-        likelihood.register_prior("noise_prior", LogNormalPrior(-1.5, 0.1), "noise")
+        likelihood.register_prior("noise_prior", UniformPrior(0.05, 0.3), "noise")
 
         mll = gpytorch.mlls.ExactMarginalLogLikelihood(likelihood, gp_model)
 
